/* graphutils.c
 *
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file
 * Various utilities for graphs.
 */

#include <stdio.h>
#include <string.h>
#include "graph_config.h"

/**
 * Print a graph (adjacency list) in human readable form
 * @param[in]  f        file pointer
 * @param[in] graph    list of neighbours of all vertices, 0-based
 * @param[in] degrees  degrees of the vertices
 * @param[in] nv       number of vertices in te graph
 */
void print_graph(FILE *f, ADJ_LIST_PTR graph, unsigned int *degrees, int nv) {
    for (int i = 0; i < nv; i++) {
        fprintf(f, "%d :", i);
        for (int j = 0; j < degrees[i]; j++)
            fprintf(f, " %d", graph[i][j]);
        fprintf(f, "\n");
    }
}

/**
 * Print an adjacency matrix in human readable form
 * @param[in]  f       file pointer
 * @param[in] matrix   adjacency matrix
 * @param[in] nv       number of vertices in te graph
 */
void print_adj_matrix(FILE *f, int nv, ADJ_MATRIX_PTR matrix) {
    for(int i = 0; i < nv; i++) {
        for(int j = 0; j < nv; j++)
            fprintf(f, "%d ", matrix[i][j]);
        fprintf(f, "\n");
    }
}

/**
 * Convert an adjacency list to a char based adjacency matrix
 */
void to_adj_matrix(ADJ_LIST_PTR graph, unsigned int *degrees, int nv, ADJ_MATRIX_PTR matrix) {
    memset(matrix, 0, MAXN * MAXN * sizeof(char));
    for (int i = 0; i < nv; i++)
        for (int j = 0; j < degrees[i]; j++)
            matrix[i][graph[i][j]] = 1;
}

/**
 * Convert an adjacency list to an int based adjacency matrix
 */
void to_int_adj_matrix(ADJ_LIST_PTR graph, unsigned int *degrees, int nv, INT_ADJ_MATRIX_PTR matrix) {
    memset(matrix, 0, MAXN * MAXN * sizeof(int));
    for (int i = 0; i < nv; i++)
        for (int j = 0; j < degrees[i]; j++)
            matrix[i][graph[i][j]] = 1;
}

/**
 * Convert a binary graph representation to a char based adjacency matrix
 */
void bits_to_adj_matrix(unsigned int *graph, int nv, ADJ_MATRIX_PTR matrix) {
    memset(matrix, 0, MAXN * MAXN * sizeof(char));
    for (int i = 0; i < nv; i++) {
        for (int j = 0, bits = graph[i]; j < nv; j++, bits <<= 1) {
            matrix[i][j] = (unsigned char)((bits & 0x80000000) != 0);
        }
    }
}

/**
 * Convert a binary graph representation to a char based adjacency matrix
 */
void bits_to_adj_list(unsigned int *graph, int nv, ADJ_LIST_PTR adj_list, unsigned int *degrees) {
    for(int i = 0; i < nv; i++) {
        degrees[i] = 0;
        for (unsigned int j = 0, bits = graph[i]; j < nv; j++, bits <<= 1) {
            if((bits & 0x80000000) != 0)
                adj_list[i][degrees[i]++] = j;
        }
    }
}

/**
 * Convert a binary graph representation to an int based adjacency matrix
 */
void bits_to_int_adj_matrix(unsigned int *graph, int nv, INT_ADJ_MATRIX_PTR matrix) {
    memset(matrix, 0, MAXN * MAXN * sizeof(int));
    for (int i = 0; i < nv; i++) {
        for (int j = 0, bits = graph[i]; j < nv; j++, bits <<= 1) {
            matrix[i][j] = (unsigned int) ((bits & 0x80000000) != 0);
        }
    }
}
