/* nutgen-old.c
 *
 * Copyright (c) 2016 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * @file
 * Program which generates nut graphs only. Version of geng with
 * additional filters to only generate nut graphs.
 *
 * Usage example: ./nutgen-old -d2 -D3 -q -c 15 | wc -l
 */

#include <stdio.h>
#include "libs/pgeng.h"

#include "graph_config.h"
#include "graphutils.h"
#include "mp.h"
#include "pseudo.h"
#include "modulo.h"
#include "longmod.h"

static ADJ_MATRIX adj_matrix;
static MP_MATRIX mp_matrix;   // A
static MP_MATRIX mp_matrix_inverse; // scalar mulitple of A.inverse

int prune_print_graph(unsigned int *g, int n, int maxn) {
    if (n == maxn) {
        bits_to_adj_matrix(g, n, adj_matrix);
        print_adj_matrix(stderr, n, adj_matrix);
        fprintf(stderr, "\n");
    }
    return 0; // everything is OK
}

/*
 * Generate only nut graphs. Prunes graphs at level maxn-1 that are singular
 * and checks at level n whether the graph is a nut graph.
 *
 * Slow version which only uses multi precision
 */
int prune_singular_parent_slow(unsigned int *g, int n, int maxn) {
    if (n == maxn) {
        bits_to_old_mp_matrix(g, n, mp_matrix);
        int rank = reduce_mp_matrix(n, mp_matrix);
        if (rank == n - 1) {
            return !check_zeroless_kernel_vector_mp_matrix(n, mp_matrix);
        } else {
            return 1; // not a nut
        }
    } else if (n + 1 == maxn) {
        bits_to_old_mp_matrix(g, n, mp_matrix);
        return reduce_mp_matrix(n, mp_matrix) != n; // 0 if nonsingular
    } else {
        return 0;
    }
}

static INT_ADJ_MATRIX int_matrix;

static int pseudo_inverse_valid;

static unsigned long long counter[3];
static unsigned long long pseudo_inverse_could_decide[3];

static unsigned long long num_graphs_generated[MAXN+1];
static unsigned long long num_graphs_accepted[MAXN+1];

/*
 * Generate only nut graphs. Prunes graphs at level maxn-1 that are singular
 * and checks at level n whether the graph is a nut graph.
 *
 * Faster version which keeps track of information on the penultimate step
 * to allow faster computations in the last step
 */
int prune_singular_parent_tracked(unsigned int *g, int n, int maxn) {
    if (n == maxn) {
        if (pseudo_inverse_valid) {
            if (pseudo_norm_bit_vector(g[n - 1], n - 1) != 0) {
                pseudo_inverse_could_decide[0]++;
                return 1; // full rank, hence not a nut
            } else {
                pseudo_inverse_could_decide[1]++;
            }
        } else {
            pseudo_inverse_could_decide[2]++;
        }

        // old pseudo method
        bits_to_int_adj_matrix(g, n, int_matrix);
        if (pseudo_reduce_matrix(n, int_matrix) == n) {
            return 1; // full rank, not a nut
        }

        bits_to_old_mp_matrix(g, n, mp_matrix);
        int rank = reduce_mp_matrix(n, mp_matrix);
        if (rank == n - 1) {
            return !check_zeroless_kernel_vector_mp_matrix(n, mp_matrix);
        } else {
            return 1; // not a nut
        }

    } else if (n + 1 == maxn) {
        bits_to_int_adj_matrix(g, n, int_matrix);

        int ret_code = near_inverse_pseudo_matrix(n, int_matrix);
        if (ret_code == 1) {
            pseudo_inverse_valid = 1;
            counter[0]++;
            return 0; // nonsingular
        } else if (ret_code == -1) {
            pseudo_inverse_valid = 0;
            counter[1]++;
            return 0; // nonsingular
        } else {
            pseudo_inverse_valid = 0;
            counter[2]++;
            bits_to_old_mp_matrix(g, n, mp_matrix);
            return reduce_mp_matrix(n, mp_matrix) != n; // 0 if nonsingular
        }
    } else {
        return 0;
    }
}

/*
 * Generate only nut graphs. Same as @ref{prune_singular_parent_tracked} but uses
 * modular arithmetic for the pseudo inverse etc.
 */
int prune_singular_parent_tracked_modulo(unsigned int *g, int n, int maxn) {
    if (n == maxn) {
        if (pseudo_inverse_valid) {
            if (modulo_norm_bit_vector(g[n - 1], n - 1) != 0) {
                pseudo_inverse_could_decide[0]++;
                return 1; // full rank, hence not a nut
            } else {
                pseudo_inverse_could_decide[1]++;
            }
        } else {
            pseudo_inverse_could_decide[2]++;
        }

        bits_to_old_mp_matrix(g, n, mp_matrix);
        int rank = reduce_mp_matrix(n, mp_matrix);
        if (rank == n - 1) {
            return !check_zeroless_kernel_vector_mp_matrix(n, mp_matrix);
        } else {
            return 1; // not a nut
        }

    } else if (n + 1 == maxn) {
        bits_to_int_adj_matrix(g, n, int_matrix);

        int ret_code = inverse_modulo_matrix(n, int_matrix);
        if (ret_code == 1) {
            pseudo_inverse_valid = 1;
            counter[0]++;
            return 0; // nonsingular
        } else {
            pseudo_inverse_valid = 0;
            counter[2]++;
            bits_to_old_mp_matrix(g, n, mp_matrix);
            //Always results 1 (i.e. prune) for small orders?! (But won't help on this level)
            return reduce_mp_matrix(n, mp_matrix) != n; // 0 if nonsingular
        }
    } else {
        return 0;
    }
}

int prune_singular_parent_tracked_longmod(unsigned int *g, int n, int maxn) {
    if (n == maxn) {
//        if (longmod_nut_property(g[n - 1], n - 1)) {
        if (longmod_nut_property_with_count(g[n - 1], n - 1)) {
            return 0; // this is a nut graph
        } else {
            return 1;
        }
    } else if (n + 1 == maxn) {
        bits_to_int_adj_matrix(g, n, int_matrix);

        int ret_code = inverse_longmod_matrix(n, int_matrix);
        if (ret_code == 1) {
            pseudo_inverse_valid = 1;
            counter[0]++;
            return 0; // nonsingular
        } else {
            pseudo_inverse_valid = 0;
            counter[2]++;
            return 1; // singular
        }
    } else {
        return 0;
    }
}


int prune_singular_parent_tracked_modulo_stats(unsigned int *g, int n, int maxn) {
    num_graphs_generated[n]++;
    int result = prune_singular_parent_tracked_modulo(g,  n, maxn);
    
    if(result == 0)
        num_graphs_accepted[n]++;
    
    return result;
}

/*
 * Generate only nut graphs. Prunes graphs at level maxn-1 that are singular
 * and checks at level n whether the graph is a nut graph.
 *
 * Faster version which also uses pseudorank
 */
int prune_singular_parent(unsigned int *g, int n, int maxn) {
    if (n == maxn) {
        bits_to_int_adj_matrix(g, n, int_matrix);
        if (pseudo_reduce_matrix(n, int_matrix) == n) {
            return 1; // full rank, not a nut
        } else {
            bits_to_old_mp_matrix(g, n, mp_matrix);
            int rank = reduce_mp_matrix(n, mp_matrix);
            if (rank == n - 1) {
                return !check_zeroless_kernel_vector_mp_matrix(n, mp_matrix);
            } else {
                return 1; // not a nut
            }
        }
    } else if (n + 1 == maxn) {
        bits_to_int_adj_matrix(g, n, int_matrix);
        if (pseudo_reduce_matrix(n, int_matrix) == n) {
            return 0; // nonsingular
        } else {
            bits_to_old_mp_matrix(g, n, mp_matrix);
            return reduce_mp_matrix(n, mp_matrix) != n; // 0 if nonsingular
        }
    } else {
        return 0;
    }
}

int main(int argc, char **argv) {
    if (MAXN > 32) {
        fprintf(stderr, "Currently only supports only up to 32 vertices\n");
        return 3;
    }
    if (!init_modulo()) {
        fprintf(stderr, "Incorrect prime or primitive element\n");
        return 3; //
    }
    init_mp_matrix(mp_matrix);
    fprintf(stderr, "Info: Using prune_singular_parent_tracked_longmod\n");
    int exit_code = pgeng(argc, argv, prune_singular_parent_tracked_longmod, 0);
    //int exit_code = pgeng(argc, argv, prune_singular_parent_tracked, 0);
    
    fprintf(stderr, "Printing statistics:\n");
    
    int i;
    for(i = 1; i <= MAXN; i++)
        if(num_graphs_generated[i] > 0)
            fprintf(stderr, "Nv=%d, num graphs generated: %llu, accepted: %llu (%.1f%%)\n", i, num_graphs_generated[i], num_graphs_accepted[i], (double) num_graphs_accepted[i] / num_graphs_generated[i] * 100);
    
    unsigned long long int total_counter = counter[0] + counter[1] + counter[2];
    fprintf(stderr, "Pseudo reusable: %llu (%.1f%%)\n", counter[0], (double) counter[0] / total_counter * 100);
    fprintf(stderr, "Pseudo nonreusable: %llu (%.1f%%)\n", counter[1], (double) counter[1] / total_counter * 100);
    fprintf(stderr, "Pseudo other: %llu (%.1f%%)\n", counter[2], (double) counter[2] / total_counter * 100);
    
    unsigned long long int total_counter_pseudo_inverse = pseudo_inverse_could_decide[0] + pseudo_inverse_could_decide[1] + pseudo_inverse_could_decide[2];
    fprintf(stderr, "Pseudo inverse could decide: %llu (%.1f%%)\n", pseudo_inverse_could_decide[0], (double) pseudo_inverse_could_decide[0] / total_counter_pseudo_inverse * 100);
    fprintf(stderr, "Pseudo inverse could not decide: %llu (%.1f%%)\n", pseudo_inverse_could_decide[1], (double) pseudo_inverse_could_decide[1] / total_counter_pseudo_inverse * 100);
    fprintf(stderr, "Pseudo inverse unusable: %llu (%.1f%%)\n", pseudo_inverse_could_decide[2], (double) pseudo_inverse_could_decide[2] / total_counter_pseudo_inverse * 100);

    longmod_print_counts(MAXN); // todo: actual size

    return exit_code;
}
